package be.iminds.ilabt.jfed.experimenter_gui.gts_editor.views;

import be.iminds.ilabt.jfed.experimenter_gui.canvas.AbstractEditableCanvas;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.CanvasNode;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.NewCanvasLink;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.NewLinkTarget;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.gts.*;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ModelRspecEditor;
import be.iminds.ilabt.jfed.experimenter_gui.gts_editor.GTSModelEditor;
import be.iminds.ilabt.jfed.experimenter_gui.gts_editor.events.ShowGtsTypeEditorEvent;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.GtsConfig;
import be.iminds.ilabt.jfed.gts.model.GtsType;
import be.iminds.ilabt.jfed.gts.model.impl.*;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import com.google.common.eventbus.EventBus;
import javafx.scene.control.*;
import javafx.scene.input.MouseButton;
import javafx.scene.input.MouseEvent;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 5/21/2015
 * Time: 10:28
 */
public class EditableGTSCanvas extends AbstractEditableCanvas {
    private static final Logger LOG = LoggerFactory.getLogger(EditableGTSCanvas.class);
    private final GTSModelEditor modelEditor;
    private final FXGtsType model;
    private final FXGtsFactory gtsFactory = new FXGtsFactory();
    private final GtsTypeToCanvasBinding binding;
    private final EventBus eventBus;
    private final GtsConfig config;

    private final List<ExternalDomain> externalDomains;

    private final ContextMenu nodeContextMenu;
    private final ContextMenu linkContextMenu;
    private final ContextMenu externalDomainContextMenu;
    private final ContextMenu typeDefinitionContextMenu;
    private final ContextMenu typeInstanceContextMenu;

    public EditableGTSCanvas(GTSModelEditor modelEditor, JFedGuiConfig config, EventBus eventBus) {
        this.modelEditor = modelEditor;
        this.config = config.getGtsConfig();
        this.eventBus = eventBus;
        this.model = modelEditor.getModel();

        this.binding = new GtsTypeToCanvasBinding(model, this);

        // create context menus
        {
            this.nodeContextMenu = new ContextMenu();

            MenuItem configureNameItem =
                    new MenuItem("Change node name", GlyphUtils.createSmallGlyph(FontAwesome.Glyph.QUOTE_LEFT));
            configureNameItem.setOnAction(actionEvent -> {
                assert (getSelectionProvider().getSelectedObject() instanceof GtsCanvasNode);
                GtsCanvasNode gtsCanvasNode = (GtsCanvasNode) getSelectionProvider().getSelectedObject();

                onConfigureNodeNamePressed(gtsCanvasNode.getNode());
            });
            nodeContextMenu.getItems().add(configureNameItem);


            MenuItem configureItem =
                    new MenuItem("Configure node", GlyphUtils.createSmallGlyph(FontAwesome.Glyph.PENCIL));
            configureItem.setOnAction(actionEvent -> {
                assert (getSelectionProvider().getSelectedObject() instanceof GtsCanvasNode);
                GtsCanvasNode canvasNode = (GtsCanvasNode) getSelectionProvider().getSelectedObject();

                onConfigureNodePressed(canvasNode);
            });
            nodeContextMenu.getItems().add(configureItem);

            MenuItem deleteItem = new MenuItem("Delete", GlyphUtils.createSmallGlyph(FontAwesome.Glyph.TIMES));
            deleteItem.setOnAction(actionEvent -> {
                assert (getSelectionProvider().getSelectedObject() instanceof GtsCanvasNode);
                GtsCanvasNode rspecCanvasNode = (GtsCanvasNode) getSelectionProvider().getSelectedObject();

                onDeleteNodePressed(rspecCanvasNode);
            });
            nodeContextMenu.getItems().add(deleteItem);
        }

        //contextMenu for links
        {
            this.linkContextMenu = new ContextMenu();

            // was disabled as FXRspecLink.LinkSetting is currently not supported in practice, but it seems to work for VM's now
            MenuItem configureLinkItem =
                    new MenuItem("Configure link", GlyphUtils.createSmallGlyph(FontAwesome.Glyph.PENCIL));
            configureLinkItem.setOnAction(actionEvent -> {
                GtsCanvasLink gtsCanvasLink = (GtsCanvasLink) getSelectionProvider().getSelectedObject();

                onConfigureLinkPressed(gtsCanvasLink.getGtsLink());
            });
            linkContextMenu.getItems().add(configureLinkItem);

            MenuItem linkDeleteItem = new MenuItem("Delete", GlyphUtils.createSmallGlyph(FontAwesome.Glyph.TIMES));
            linkDeleteItem.setOnAction(actionEvent -> {
                GtsCanvasLink gtsCanvasLink = (GtsCanvasLink) getSelectionProvider().getSelectedObject();

                onDeleteLinkPressed(gtsCanvasLink);
            });
            linkContextMenu.getItems().add(linkDeleteItem);
        }

        {
            this.externalDomainContextMenu = new ContextMenu();

            MenuItem configureDomainItem =
                    new MenuItem("Configure External Domain", GlyphUtils.createSmallGlyph(FontAwesome.Glyph.PENCIL));
            configureDomainItem.setOnAction(actionEvent -> {
                assert (getSelectionProvider().getSelectedObject() instanceof GtsCanvasNode);
                GtsCanvasNode canvasNode =
                        (GtsCanvasNode) getSelectionProvider().getSelectedObject();

                assert canvasNode.getNode() instanceof FXGtsExternalNetwork;

                onConfigureExternalDomainPressed((FXGtsExternalNetwork) canvasNode.getNode());
            });
            externalDomainContextMenu.getItems().add(configureDomainItem);

            MenuItem deleteDomainItem = new MenuItem("Delete", GlyphUtils.createSmallGlyph(FontAwesome.Glyph.TIMES));
            deleteDomainItem.setOnAction(actionEvent -> {
                assert (getSelectionProvider().getSelectedObject() instanceof GtsCanvasNode);
                GtsCanvasNode rspecCanvasNode = (GtsCanvasNode) getSelectionProvider().getSelectedObject();

                onDeleteNodePressed(rspecCanvasNode);
            });
            externalDomainContextMenu.getItems().add(deleteDomainItem);
        }

        {
            this.typeDefinitionContextMenu = new ContextMenu();

            MenuItem instanciateMenuItem = new MenuItem("Instantiate", GlyphUtils.createSmallGlyph('\uf24d'));
            instanciateMenuItem.setOnAction(actionEvent -> {
                assert (getSelectionProvider().getSelectedObject() instanceof GtsCanvasNode);
                GtsCanvasTypeDefinition rspecCanvasNode =
                        (GtsCanvasTypeDefinition) getSelectionProvider().getSelectedObject();

                onInstantiateTypePressed(rspecCanvasNode);
            });
            typeDefinitionContextMenu.getItems().add(instanciateMenuItem);


            MenuItem deleteTypeMenuItem = new MenuItem("Delete", GlyphUtils.createSmallGlyph(FontAwesome.Glyph.TIMES));
            deleteTypeMenuItem.setOnAction(actionEvent -> {
                assert (getSelectionProvider().getSelectedObject() instanceof GtsCanvasNode);
                GtsCanvasTypeDefinition rspecCanvasNode =
                        (GtsCanvasTypeDefinition) getSelectionProvider().getSelectedObject();

                onDeleteTypeDefinitionPressed(rspecCanvasNode);
            });
            typeDefinitionContextMenu.getItems().add(deleteTypeMenuItem);
        }

        {
            this.typeInstanceContextMenu = new ContextMenu();

            MenuItem deleteTypeMenuItem = new MenuItem("Delete", GlyphUtils.createSmallGlyph(FontAwesome.Glyph.TIMES));
            deleteTypeMenuItem.setOnAction(actionEvent -> {
                assert (getSelectionProvider().getSelectedObject() instanceof GtsCanvasNode);
                GtsCanvasTypeInstance rspecCanvasNode =
                        (GtsCanvasTypeInstance) getSelectionProvider().getSelectedObject();

                onDeleteTypeInstancePressed(rspecCanvasNode);
            });
            typeInstanceContextMenu.getItems().add(deleteTypeMenuItem);
        }

        this.externalDomains =
                this.config.getExternalDomains().stream()
                        .map(ed -> new ExternalDomain(ed.getLocation(), ed.getPort(), ed.getName(), ed.getDescription()))
                        .collect(Collectors.toList());

    }

    private void onConfigureNodeNamePressed(FXGtsNode gtsNode) {

        TextInputDialog dialog = new TextInputDialog(gtsNode.getId());
        dialog.setHeaderText("Edit node name");
        dialog.setContentText("New node name:");

        dialog.showAndWait().ifPresent(newName -> {
            if (ModelRspecEditor.isValidNodeName(newName)) {
                gtsNode.setId(newName);
            } else {
                new Alert(Alert.AlertType.ERROR, "The given name is invalid.").showAndWait();
            }
        });
    }

    public boolean onConfigureExternalDomainPressed(FXGtsExternalNetwork externalNetwork) {


        String portId = !externalNetwork.getPorts().isEmpty() ?
                externalNetwork.getPorts().values().iterator().next().getId() : null;

        ExternalDomain currentDomain =
                externalDomains.stream().filter(domain -> domain.getLocation().equals(externalNetwork.getLocation())
                        && domain.getPortId().equals(portId)).findFirst()
                        .orElse(null);


        ChoiceDialog<ExternalDomain> dialog = new ChoiceDialog<>(currentDomain, externalDomains);

        dialog.setHeaderText("Select an external domain");
        dialog.setTitle("Select an external domain");

        Optional<ExternalDomain> result = dialog.showAndWait();
        if (result.isPresent()) {
            ExternalDomain externalDomain = result.get();
            externalNetwork.setLocation(externalDomain.getLocation());
            externalNetwork.setRequiredPortName(externalDomain.getPortId());

            if (!externalNetwork.getPorts().isEmpty()) {
                //change the port name

                FXGtsPort existingPort = externalNetwork.getPorts().values().iterator().next();

                externalNetwork.getPorts().remove(existingPort.getId());
                existingPort.setId(externalDomain.getPortId());
                externalNetwork.getPorts().put(existingPort.getId(), existingPort);
            }


            externalNetwork.getPorts().clear();
            FXGtsPort port = gtsFactory.createGtsPort();
            port.setId(externalDomain.getPortId());
            externalNetwork.addPort(port);
            return true;
        } else
            return false;
    }

    public GTSModelEditor getModelEditor() {
        return modelEditor;
    }

    private static class ExternalDomain {
        private final String location;
        private final String portId;
        private final String friendlyName;
        private final String description;

        public ExternalDomain(String location, String portId, String friendlyName, String description) {
            this.friendlyName = friendlyName;
            this.location = location;
            this.portId = portId;
            this.description = description;
        }

        public String getFriendlyName() {
            return friendlyName;
        }

        public String getDescription() {
            return description;
        }

        public String getLocation() {
            return location;
        }

        public String getPortId() {
            return portId;
        }

        public String toString() {
            return friendlyName + " (" + description + ")";
        }
    }

    private void onConfigureLinkPressed(FXGtsLink gtsLink) {
        TextInputDialog dialog = new TextInputDialog("10");
        dialog.setTitle("Link settings");
        dialog.setHeaderText("Link Settings");
        dialog.setContentText("Capacity [in Mbps]:");

        dialog.showAndWait().ifPresent(answer -> {
            boolean success = true;

            try {
                int value = Integer.parseInt(answer);
                if (value >= 1 && value <= 10_000) {
                    gtsLink.setCapacity(value);
                } else {
                    success = false;
                }
            } catch (NumberFormatException ex) {
                success = false;
            }

            if (!success) {
                new Alert(Alert.AlertType.ERROR, "Invalid capacity was entered! Only integer values between 1 and 10000 are allowed!").showAndWait();
            }
        });

    }

    private void onConfigureNodePressed(GtsCanvasNode gtsCanvasNode) {
        ChoiceDialog<String> dialog = new ChoiceDialog<>(gtsCanvasNode.getNode().getLocation(), config.getLocations());

        dialog.setContentText("Select the location for this host:");
        dialog.showAndWait().ifPresent(location -> gtsCanvasNode.getNode().setLocation(location));

    }

    @Override
    protected void deleteSelectedItem() {
        if (getSelectionProvider().getSelectedObject() instanceof GtsCanvasNode) {
            onDeleteNodePressed((GtsCanvasNode) getSelectionProvider().getSelectedObject());

        } else if (getSelectionProvider().getSelectedObject() instanceof GtsCanvasLink) {
            onDeleteLinkPressed((GtsCanvasLink) getSelectionProvider().getSelectedObject());
        }
    }

    @Override
    public boolean isLegalLink(NewCanvasLink newCanvasLink) {
        if (!newCanvasLink.isLinkComplete()) {
            LOG.trace("Requested link is not complete");
            return false;
        }

        if (!isValidForLink(newCanvasLink.getOrigin()) || !(isValidForLink(newCanvasLink.getTarget())))
            return false;

        if (newCanvasLink.getOrigin() instanceof GtsCanvasPort && newCanvasLink.getTarget() instanceof GtsCanvasPort) {
            LOG.info("Cannot link two ports together!");
            return false;
        }

        return true;
    }

    private boolean isValidForLink(NewLinkTarget newLinkTarget) {
        if (newLinkTarget instanceof GtsCanvasNode) {
            return isValidNodeForNewLink(((GtsCanvasNode) newLinkTarget).getNode());
        } else if (newLinkTarget instanceof GtsCanvasPort) {
            return isValidPortForNewLink(((GtsCanvasPort) newLinkTarget).getPort());
        } else {
            LOG.warn("Unknown object is checked for validity of link: {}", newLinkTarget);
            return true;
        }
    }

    private boolean isValidNodeForNewLink(FXGtsNode node) {
        if (node instanceof FXGtsExternalNetwork) {
            if (!node.getPorts().isEmpty())
                return model.getAdjacencyInfo(node.getPorts().values().iterator().next()) == null;
            else
                return true;
        }

        return true;
    }

    private boolean isValidPortForNewLink(FXGtsPort port) {
        //do not allow ports that already have an adjacency
        if (model.findAdjacency(null, port.getId()) != null) {
            LOG.trace("Port already has an adjacency");
            return false;
        }
        return true;
    }

    private void onDeleteNodePressed(GtsCanvasNode gtsCanvasNode) {
        modelEditor.removeNode(gtsCanvasNode.getNode());
        if (getSelectionProvider().getSelectedObject() == gtsCanvasNode) {
            getSelectionProvider().clearAllSelections();
        }

    }

    private boolean onInstantiateTypePressed(GtsCanvasTypeDefinition gtsCanvasTypeDefinition) {

        modelEditor.addTypeInstance(gtsCanvasTypeDefinition.getType());

        return true;
    }

    private boolean onDeleteTypeDefinitionPressed(GtsCanvasTypeDefinition gtsCanvasNode) {
        final FXGtsType type = gtsCanvasNode.getType();
        assert type.isDefinition();
        if (type.isDefinition()) {
            int instancesCount = modelEditor.countTypeInstances(type);
            if (instancesCount > 0) {
                new Alert(Alert.AlertType.WARNING,
                        String.format("Type '%s' has %d instance(s), and cannot be deleted. " +
                                        "Please delete all instances before deleting the declaration",
                                type.getName(), instancesCount)).showAndWait();

                return false;
            }
        }


        modelEditor.removeType(type);
        if (getSelectionProvider().getSelectedObject() == gtsCanvasNode) {
            getSelectionProvider().clearAllSelections();
        }

        return true;

    }

    private void onDeleteTypeInstancePressed(GtsCanvasTypeInstance gtsCanvasNode) {
        final FXGtsType type = gtsCanvasNode.getType();
        assert !type.isDefinition();

        modelEditor.removeType(type);
        if (getSelectionProvider().getSelectedObject() == gtsCanvasNode) {
            getSelectionProvider().clearAllSelections();
        }
    }

    private void onDeleteLinkPressed(GtsCanvasLink gtsCanvasLink) {
        modelEditor.removeLink(gtsCanvasLink.getGtsLink());
        if (getSelectionProvider().getSelectedObject() == gtsCanvasLink) {
            getSelectionProvider().clearAllSelections();
        }
    }

    @Override
    protected boolean processNewLinkRequest(NewCanvasLink newCanvasLink) {
        LOG.trace("Received new link request between {} AND {}", newCanvasLink.getOrigin(),
                newCanvasLink.getTarget());

        //delegate the real link-creation to ModelRspecEditor
        LOG.trace("Received link request is OK!");

        if (newCanvasLink.getOrigin() instanceof GtsCanvasNode && newCanvasLink.getTarget() instanceof GtsCanvasNode) {
            final FXGtsNode originNode = ((GtsCanvasNode) newCanvasLink.getOrigin()).getNode();

            final FXGtsNode targetNode = ((GtsCanvasNode) newCanvasLink.getTarget()).getNode();
            FXGtsLink newLink = modelEditor.addLink(originNode, targetNode);

            setStatus("Created a new link '" + newLink.getId() + "'");
            return true;
        } else if (newCanvasLink.getOrigin() instanceof GtsCanvasNode) {
            assert newCanvasLink.getTarget() instanceof GtsCanvasPort;

            final FXGtsNode node = ((GtsCanvasNode) newCanvasLink.getOrigin()).getNode();
            final FXGtsPort port = ((GtsCanvasPort) newCanvasLink.getTarget()).getPort();
            GtsType.Adjacency adjacency = modelEditor.addAdjacency(node, port);
            setStatus("Created an adjacency between '" + node.getId() + "' and '" + port.getId() + "'");
            return true;
        } else if (newCanvasLink.getTarget() instanceof GtsCanvasNode) {
            assert newCanvasLink.getOrigin() instanceof GtsCanvasPort;

            final FXGtsNode node = ((GtsCanvasNode) newCanvasLink.getTarget()).getNode();
            final FXGtsPort port = ((GtsCanvasPort) newCanvasLink.getOrigin()).getPort();
            GtsType.Adjacency adjacency = modelEditor.addAdjacency(node, port);
            setStatus("Created an adjacency between '" + node.getId() + "' and '" + port.getId() + "'");
            return true;
        } else {
            return false;
        }
    }

    @Override
    protected CanvasNode addCopyOfNodeToCanvas(CanvasNode originalRspecCanvasNode) {
        //TODO implement copy
        return null;
    }

    @Override
    public void canvasItemOnMouseClicked(MouseEvent mouseEvent) {
        super.canvasItemOnMouseClicked(mouseEvent);

        if (mouseEvent.getSource() instanceof GtsCanvasNode) {
            GtsCanvasNode canvasNode = (GtsCanvasNode) mouseEvent.getSource();
            getSelectionProvider().setSelectedObject(canvasNode);

            if (canvasNode.getNode() instanceof FXGtsExternalNetwork) {
                if (mouseEvent.getButton() == MouseButton.PRIMARY && mouseEvent.getClickCount() == 2) {
                    onConfigureExternalDomainPressed((FXGtsExternalNetwork) canvasNode.getNode());

                } else if (mouseEvent.getButton() != MouseButton.PRIMARY || mouseEvent.isControlDown()) {
                    //show context menu
                    externalDomainContextMenu.show(canvasNode, mouseEvent.getScreenX(), mouseEvent.getScreenY());
                }
            } else if (canvasNode.getNode() instanceof FXGtsHost) {
                if (mouseEvent.getButton() == MouseButton.PRIMARY && mouseEvent.getClickCount() == 2) {
                    onConfigureNodePressed(canvasNode);

                } else if (mouseEvent.getButton() != MouseButton.PRIMARY || mouseEvent.isControlDown()) {
                    //show context menu
                    nodeContextMenu.show(canvasNode, mouseEvent.getScreenX(), mouseEvent.getScreenY());
                }
            } else
                LOG.debug("No action defined for click on {}", canvasNode.getNode().getClass().getName());

        }

        if (mouseEvent.getSource() instanceof GtsCanvasLink) {
            final GtsCanvasLink link = (GtsCanvasLink) mouseEvent.getSource();
            getSelectionProvider().setSelectedObject(link);

            if (mouseEvent.getButton() == MouseButton.PRIMARY && mouseEvent.getClickCount() == 2) {
                onConfigureLinkPressed(link.getGtsLink());
            } else if (mouseEvent.getButton() != MouseButton.PRIMARY || mouseEvent.isControlDown()) {
                //show context menu
                linkContextMenu.show(link, mouseEvent.getScreenX(), mouseEvent.getScreenY());
            }
        }

        if (mouseEvent.getSource() instanceof GtsCanvasType) {
            final GtsCanvasType canvasType = (GtsCanvasType) mouseEvent.getSource();
            getSelectionProvider().setSelectedObject(canvasType);

            if (mouseEvent.getButton() == MouseButton.PRIMARY && mouseEvent.getClickCount() == 2) {
                if (canvasType.getType().isDefinition()) {
                    onEditGTSTypePressed(canvasType);
                } else {
                    //FIXME Implement instance editor
                }
            } else if (mouseEvent.getButton() != MouseButton.PRIMARY || mouseEvent.isControlDown()) {
                //show context menu
                if (canvasType.getType().isDefinition()) {
                    typeDefinitionContextMenu.show(canvasType, mouseEvent.getScreenX(), mouseEvent.getScreenY());
                } else {
                    typeInstanceContextMenu.show(canvasType, mouseEvent.getScreenX(), mouseEvent.getScreenY());
                }

            }
        }
    }

    private void onEditGTSTypePressed(GtsCanvasType gtsCanvasType) {
        final FXGtsType type = gtsCanvasType.getType();

        eventBus.post(new ShowGtsTypeEditorEvent(modelEditor, type));
    }
}
