package be.iminds.ilabt.jfed.highlevel.jobs.states;

import be.iminds.ilabt.jfed.espec.model.AnsibleGalaxySpec;
import be.iminds.ilabt.jfed.espec.model.AnsibleHostSpec;
import be.iminds.ilabt.jfed.espec.model.AnsiblePlaybookSpec;
import be.iminds.ilabt.jfed.espec.model.AnsibleSpec;
import be.iminds.ilabt.jfed.espec.util.ESpecLogger;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.jobs.*;
import be.iminds.ilabt.jfed.highlevel.util.ByteArraySshjFile;
import be.iminds.ilabt.jfed.highlevel.util.ExternalFileUtil;
import be.iminds.ilabt.jfed.highlevel.util.StringMemorySourceFile;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.rspec.model.AnsibleService;
import be.iminds.ilabt.jfed.rspec.model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.ModelRspec;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import javafx.application.Platform;
import javafx.beans.property.SimpleStringProperty;
import javafx.beans.property.StringProperty;
import javafx.util.Pair;
import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.connection.ConnectionException;
import net.schmizz.sshj.connection.channel.direct.Session;
import net.schmizz.sshj.transport.TransportException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.event.Level;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.*;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 5/27/2016
 * Time: 11:51
 */
public class ExperimentSpecificationServicesAnsibleState extends SlicedState<ExperimentSpecificationServicesAnsibleState.ExperimentSpecificationServicesAnsibleStateSlice> {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentSpecificationServicesAnsibleState.class);

    @Nonnull private final SetupSoftwareExperimentJob job;
    @Nonnull private final Experiment experiment;
    @Nonnull private final GeniUserProvider geniUserProvider;
    @Nonnull private final ExperimentSpecificationServicesDirsAndUploadsState parent;
    @Nonnull private final AnsibleSpec ansibleSpec;

    @Nonnull private final ESpecLogger eSpecLogger;

    private final boolean useRemoteAnsibleNode;
    @Nonnull private final String ansibleNodeClientId;
    @Nonnull private final RspecNode ansibleNode;


    @Nonnull private final List<ExperimentSpecificationServicesAnsibleStateSlice> slices;

    protected ExperimentSpecificationServicesAnsibleState(@Nonnull SetupSoftwareExperimentJob job,
                                                          @Nonnull GeniUserProvider geniUserProvider,
                                                          @Nonnull ExperimentSpecificationServicesDirsAndUploadsState parent,
                                                          @Nonnull AnsibleSpec ansibleSpec) {
        super("Execute ansible services");
        this.job = job;
        this.geniUserProvider = geniUserProvider;
        this.parent = parent;
        this.ansibleSpec = ansibleSpec;
        this.experiment = job.getExperiment();
        this.eSpecLogger = experiment.getExperimentSpecificationLogger();
        experiment.requireSlice();

        /////////////

        this.useRemoteAnsibleNode = ansibleSpec.getAnsibleHostSpec().getHostType() != AnsibleHostSpec.HostType.LOCAL;

        if (!useRemoteAnsibleNode) {
            throw new RuntimeException("LOCAL ansible execute is not yet implemented.");
        } else {
            if (ansibleSpec.getAnsibleHostSpec().getNodeName() == null)
                throw new IllegalStateException("ansible node name should be set");
        }

        assert experiment.getNewRequestRspecSource() != null;
        ModelRspec request = experiment.getNewRequestRspecSource() == null ?
                null :
                experiment.getNewRequestRspecSource().getModelRspec();

        if (request == null) {
            String err = "Cannot find ansible node because parsing request rspec failed";
            LOG.error(err);
            throw new RuntimeException(err);
        }

        RspecNode ansibleNode = request.getNodeByClientId(ansibleSpec.getAnsibleHostSpec().getNodeName());
        if (ansibleNode == null || ansibleNode.getClientId() == null) {
            String err = "Cannot find ansible node \""+ansibleSpec.getAnsibleHostSpec().getNodeName()+"\" in request rspec";
            LOG.error(err);
            throw new RuntimeException(err);
        }
        this.ansibleNode = ansibleNode;
        this.ansibleNodeClientId = ansibleNode.getClientId();

        this.slices = Collections.singletonList(
                new ExperimentSpecificationServicesAnsibleStateSlice(job, ansibleNode, ansibleSpec)
        );
    }

    @Override
    public Collection<ExperimentSpecificationServicesAnsibleStateSlice> getSlices() {
        return slices;
    }

    public class ExperimentSpecificationServicesAnsibleStateSlice extends StateSlice {
        @Nonnull  protected final SetupSoftwareExperimentJob setupSoftwareExperimentJob;
        @Nonnull protected final RspecNode ansibleNode;
        @Nonnull protected final AnsibleSpec ansibleSpec;

        @Nonnull private String galaxyCommand;
        @Nonnull private String playbookCommand;

        //only initialised when there is an SSH connection (relies on homedir location)
        private String ansibleDirPath;
        private String ansibleInventoryPath;
        private String ansibleInventoryBasename;
        private String ansibleSshConfigPath;
        private String ansibleSshConfigBasename;

        public ExperimentSpecificationServicesAnsibleStateSlice(@Nonnull SetupSoftwareExperimentJob job,
                                                                @Nonnull RspecNode ansibleNode,
                                                                @Nonnull AnsibleSpec ansibleSpec) {
            super(job);
            this.setupSoftwareExperimentJob = job;
            this.ansibleNode = ansibleNode;
            this.ansibleSpec = ansibleSpec;

            this.galaxyCommand = ansibleSpec.getAnsibleHostSpec().getGalaxyCommand() == null ? "" : ansibleSpec.getAnsibleHostSpec().getGalaxyCommand();
            this.playbookCommand = ansibleSpec.getAnsibleHostSpec().getPlaybookCommand() == null ? "" : ansibleSpec.getAnsibleHostSpec().getPlaybookCommand();
        }

        @Override
        public String getName() {
            return "Ansible on " + ansibleNode.getClientId();
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            SSHClient ssh;
            try {
                ssh = setupSoftwareExperimentJob.getSSHClient(ansibleNode, this);
            } catch (SetupSoftwareExperimentJob.SshException e) {
                LOG.warn("Aborting ExecuteAnsibleServiceStateSlice.statefulRun due to SSH comnection problem", e);
                return ExperimentTaskStatus.FAILED;
            }
            assert ssh != null;
            ExternalFileUtil externalFileUtil = new ExternalFileUtil(ssh);

            this.ansibleInventoryBasename = "ansible-hosts";
            this.ansibleSshConfigBasename = "ssh-config";
            this.ansibleDirPath = assureEndWithSlash(externalFileUtil.getHomeDir())+"ansible/";  //TODO: make configurable
            this.ansibleInventoryPath = assureEndWithSlash(ansibleDirPath)+ansibleInventoryBasename; //TODO: make configurable
            this.ansibleSshConfigPath = assureEndWithSlash(ansibleDirPath)+ansibleSshConfigBasename; //TODO: make configurable

            eSpecLogger.firePreAnsible();

            PrepareAnsibleDirState prepareAnsibleDirState = new PrepareAnsibleDirState(ssh);
            setAndRunState(prepareAnsibleDirState);

            ExperimentTaskStatus worstStatus = prepareAnsibleDirState.getStatus();

            for (AnsibleGalaxySpec gs : ansibleSpec.getAnsibleGalaxySpecs()) {
                RunGalaxyState executeGalaxyState = new RunGalaxyState(gs, ssh);
                setAndRunState(executeGalaxyState);
                worstStatus = ExperimentTaskStatus.worstOf(worstStatus, executeGalaxyState.getStatus());
            }

            for (AnsiblePlaybookSpec pbs : ansibleSpec.getAnsiblePlaybookSpecs()) {
                ExecutePlayBookState executePlayBookState = new ExecutePlayBookState(pbs, ssh);
                setAndRunState(executePlayBookState);
                worstStatus = ExperimentTaskStatus.worstOf(worstStatus, executePlayBookState.getStatus());
            }

            setupSoftwareExperimentJob.closeSSH(ansibleNode);

            eSpecLogger.firePostAnsibleAll(worstStatus == ExperimentTaskStatus.SUCCESS);

            return worstStatus;
        }

        public class PrepareAnsibleDirState extends State {
            @Nonnull private final ModelRspec request;
            @Nonnull private final ModelRspec manifest;

            @Nonnull private final SSHClient ssh;

            protected PrepareAnsibleDirState(@Nonnull SSHClient ssh) {
                super("Preparing ansible dir '"+ansibleDirPath+"'");

                this.ssh = ssh;

                //initialize request and manifest-model
                assert experiment.getNewRequestRspecSource() != null;
                ModelRspec request = experiment.getNewRequestRspecSource() == null ?
                        null :
                        experiment.getNewRequestRspecSource().getModelRspec();

                if (request == null) {
                    LOG.error("Cannot create inventory file because parsing request rspec failed");
                    throw new RuntimeException("Cannot create inventory file because parsing request rspec failed");
                }

                this.request = request;


                assert experiment.getSliceOrNull() != null;
                ModelRspec manifest =
                        experiment.getSliceOrNull() == null || experiment.getSlice().getManifestRspec() == null ?
                                null :
                                experiment.getSliceOrNull().getManifestRspec().getModelRspec();

                if (manifest == null) {
                    LOG.error("Cannot create inventory file because parsing manifest rspec failed");
                    throw new RuntimeException("Cannot create inventory file because parsing manifest rspec failed");
                }
                this.manifest = manifest;

                if (experiment.getSliceOrNull().getManifestRspec().getStringRspec() == null){
                    LOG.error("Cannot create inventory file because parsing manifest rspec failed");
                    throw new RuntimeException("Cannot create inventory file because parsing manifest rspec failed");
                }
            }

            public String createInventoryFile() {
                //construct a multimap containing the nodes grouped per ansible-group

                Multimap<String, RspecNode> nodesPerGroup = HashMultimap.create();
                for (RspecNode node : request.getNodes()) {
                    if (!node.getAnsibleGroups().isEmpty()) { //TODO: add a way to specify ansible groups to AnsibleSpec
                        node.getAnsibleGroups().forEach(groupName -> nodesPerGroup.put(groupName, node));
                    } else {
                        nodesPerGroup.put(null, node);
                    }
                }

                //now write the inventory file
                StringBuilder inventory = new StringBuilder();

                //first write the groupless nodes
                if (nodesPerGroup.containsKey(null)) {
                    Collection<RspecNode> grouplessNodes = nodesPerGroup.removeAll(null);
                    grouplessNodes.forEach(node -> writeNodeToInventory(node, inventory));
                    inventory.append('\n');
                }

                //now write all groups

                for (String groupName : nodesPerGroup.keySet()) {
                    inventory.append("[").append(groupName).append("]\n");
                    nodesPerGroup.get(groupName).forEach(node -> writeNodeToInventory(node, inventory));
                    inventory.append('\n');
                }

                return inventory.toString();
            }

            private void writeNodeToInventory(RspecNode node, StringBuilder inventory) {
                final BasicStringRspec.LoginService loginService =
                        setupSoftwareExperimentJob.getBestNodeLoginFinder()
                                .findBestLogin(node.getUniqueId());

                if (loginService != null) {
                    inventory.append(node.getClientId())
                            .append("\tansible_host=").append(loginService.getHostname());

                    if (loginService.getPort() != 22) {
                        inventory.append("\tansible_port=").append(loginService.getPort());
                    }

                    if (!Objects.equals(loginService.getUsername(),
                            geniUserProvider.getLoggedInGeniUser().getUserUrn().getResourceName())) {
                        inventory.append("\tansible_user=").append(loginService.getUsername());
                    }
                    inventory.append('\n');
                }
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                try {
                    //create inventory file
                    String inventory = createInventoryFile();

                    ssh.newSCPFileTransfer().upload(
                            new StringMemorySourceFile(ansibleInventoryBasename, inventory),
                            ansibleInventoryPath);
                } catch (IOException ex) {
                    LOG.error("Error while uploading inventory-file to {} on {}", ansibleInventoryPath, ansibleNodeClientId, ex);
                    updateMessage(String.format("Error while uploading inventory-file to %s on %s: %s", ansibleInventoryPath, ansibleNodeClientId, ex.getMessage()));
                    return ExperimentTaskStatus.FAILED;
                }

                return ExperimentTaskStatus.SUCCESS;
            }
        }

        public class RunGalaxyState extends State {
            @Nonnull private final AnsibleGalaxySpec galaxySpec;
            @Nonnull private final SSHClient ssh;

            @Nonnull private final StringProperty galaxyInputStream, galaxyErrorStream;

            protected RunGalaxyState(@Nonnull AnsibleGalaxySpec galaxySpec, @Nonnull SSHClient ssh) {
                super("Running galaxy requirements '" + galaxySpec.getSource().getBasename() + "'");

                this.galaxySpec = galaxySpec;
                this.ssh = ssh;

                this.galaxyInputStream = new SimpleStringProperty("");
                this.galaxyErrorStream = new SimpleStringProperty("");
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                try {
                    String remotePath = parent.getRemotePath(galaxySpec);
                    if (remotePath == null) {
                        String err = "Did not find remote path of ansible galaxy requirements. Bug in ansible step?";
                        LOG.error(err);
                        updateMessage(err);
                        throw new RuntimeException(err);
                        //                    return ExperimentTaskStatus.FAILED;
                    }

                    //isntall ansible galaxy requirements
                    String galaxyCommandString = String.format("cd %s && %s install -r %s -p roles",
                            ansibleDirPath, galaxyCommand, remotePath);
                    LOG.debug("Executing galaxy command: "+galaxyCommandString);
                    Session.Command c =
                            ssh.startSession().exec(galaxyCommandString);

                    new InputStreamToStringPropertyReaderThread(c.getInputStream(), galaxyInputStream, Level.INFO, "GALAXY-OUTPUT: ").start();
                    new InputStreamToStringPropertyReaderThread(c.getErrorStream(), galaxyErrorStream, Level.WARN, "GALAXY-ERROR: ").start();

                    c.join();

                    if (c.getExitStatus() != 0) {
                        updateMessage("ansible-galaxy execution was unsuccessful. Aborting");
                        return ExperimentTaskStatus.FAILED;
                    }
                } catch (TransportException | ConnectionException ex) {
                    throw new JFedException("SSH-exception while executing ansible: " + ex.getMessage(), ex);
                }

                return ExperimentTaskStatus.SUCCESS;
            }

            @Nonnull
            public StringProperty getGalaxyInputStream() {
                return galaxyInputStream;
            }

            @Nonnull
            public StringProperty getGalaxyErrorStream() {
                return galaxyErrorStream;
            }
        }

        public class ExecutePlayBookState extends State {
            @Nonnull private final AnsiblePlaybookSpec playbook;
            @Nonnull private final SSHClient ssh;

            private final StringProperty playbookInputStream, playbookErrorStream;


            protected ExecutePlayBookState(@Nonnull AnsiblePlaybookSpec playbook, @Nonnull SSHClient ssh) {
                super("Executing ansible-playbook '" + playbook.getSource().getBasename() + "'");

                this.playbook = playbook;
                this.ssh = ssh;

                this.playbookInputStream = new SimpleStringProperty("");
                this.playbookErrorStream = new SimpleStringProperty("");
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                String remotePath = parent.getRemotePath(playbook);
                if (remotePath == null) {
                    eSpecLogger.firePostAnsiblePlaybook(playbook, "not found", ansibleNodeClientId, false);
                    String err = "Did not find remote path of ansible playbook. Bug in ansible step?";
                    LOG.error(err);
                    updateMessage(err);
                    throw new RuntimeException(err);
                    //                    return ExperimentTaskStatus.FAILED;
                }

                try {
                    String arguments = "";
                    arguments += remotePath;

                    if (Objects.equals(playbook.isDebug(), Boolean.TRUE))
                        arguments += " -v";

                    arguments += " -i '"+ansibleInventoryPath+"'";

                    //TODO: log ansible output to file
//                    //this must be the LAST argument, as it is a redirect and pipe
//                    if (ansibleService.getLogfile() != null)
//                        arguments += " 2>&1 | tee '"+ansibleService.getLogfile()+"'";

                    String playbookCommandString = String.format("cd %s && %s %s", ansibleDirPath, playbookCommand, arguments);
                    LOG.debug("Executing playbook command: "+playbookCommandString);
                    Session.Command c =
                            ssh.startSession().exec(playbookCommandString);

                    new InputStreamToStringPropertyReaderThread(c.getInputStream(), playbookInputStream, Level.INFO, "PLAYBOOK-OUTPUT: ").start();
                    new InputStreamToStringPropertyReaderThread(c.getErrorStream(), playbookErrorStream, Level.WARN, "PLAYBOOK-ERROR: ").start();

                    c.join();

                    if (c.getExitStatus() != 0) {
                        updateMessage("ansible-playbook execution was unsuccessful.");
                        eSpecLogger.firePostAnsiblePlaybook(playbook, remotePath, ansibleNodeClientId, false);
                        return ExperimentTaskStatus.FAILED;
                    } else {
                        eSpecLogger.firePostAnsiblePlaybook(playbook, remotePath, ansibleNodeClientId, true);
                        return ExperimentTaskStatus.SUCCESS;
                    }
                } catch (TransportException | ConnectionException ex) {
                    eSpecLogger.firePostAnsiblePlaybook(playbook, remotePath, ansibleNodeClientId, false);
                    throw new JFedException("SSH-exception while executing ansible: " + ex.getMessage(), ex);
                }
            }

            public StringProperty getPlaybookInputStream() {
                return playbookInputStream;
            }

            public StringProperty getPlaybookErrorStream() {
                return playbookErrorStream;
            }
        }
    }

    private static class InputStreamToStringPropertyReaderThread extends Thread {

        private final InputStream inputStream;
        private final StringProperty output;

        private final String logPrefix;
        private final LogFunction logFunction;


        private InputStreamToStringPropertyReaderThread(InputStream inputStream, StringProperty output,
                                                        @Nullable Level logLevel, String logPrefix) {
            this.inputStream = inputStream;
            this.output = output;
            this.logPrefix = logPrefix;

            if (logLevel != null) {
                switch (logLevel) {
                    case TRACE:
                        logFunction = LOG::trace;
                        break;
                    case DEBUG:
                        logFunction = LOG::debug;
                        break;
                    case INFO:
                        logFunction = LOG::info;
                        break;
                    case WARN:
                        logFunction = LOG::warn;
                        break;
                    case ERROR:
                        logFunction = LOG::error;
                        break;
                    default:
                        logFunction = null;
                }

            } else
                logFunction = null;
        }

        @Override
        public void run() {

            try {
                BufferedReader stdInput = new BufferedReader(new InputStreamReader(inputStream));

                String s;
                while ((s = stdInput.readLine()) != null) {
                    System.out.println(s);
                    String str = s;

                    if (logFunction != null) {
                        logFunction.log(logPrefix + str);
                    }

                    Platform.runLater(() -> output.setValue(output.get() + "\n" + str));

                }

                inputStream.close();
            } catch (IOException ex) {
                LOG.error("Error while reading from InputStream to StringProperty", ex);
            }

        }

        private interface LogFunction {
            void log(String text);
        }
    }

    @Nonnull
    private static String assureEndWithSlash(@Nonnull String p) {
        if (p.endsWith("/")) return p;
        return p + "/";
    }

    @Nonnull
    private static String assureStartWithSlash(@Nonnull String p) {
        if (p.startsWith("/")) return p;
        return "/" + p;
    }
}
